import fsspec
import json 
import os 
from tqdm.notebook import tqdm 
import shutil
import pandas as pd 


def rm(path, bucket='dmitrij',
       endpoint="https://8a6ab2cee54f34a71f5a8d99e92da2d2.r2.cloudflarestorage.com",
       verbose=True):

    fs = fsspec.filesystem(
        "s3",
        profile="r2",
        client_kwargs={"endpoint_url": endpoint, "use_ssl": True},
        config_kwargs={"s3": {"addressing_style": "path"}}
    )
    target = f"s3://{bucket}/{path}"
    if verbose:
        print(f"Deleting {target}...")
    fs.rm(target)


def done(fasta_folder='000',
         structure_folder='',
         bucket='dmitrij',
         endpoint="https://8a6ab2cee54f34a71f5a8d99e92da2d2.r2.cloudflarestorage.com",
         load_local_cache=True,
         load_s3_cache=True,
         verbose=True,
         pbar=True,
        ):

    fs = fsspec.filesystem(
            "s3",
            profile="r2",
            client_kwargs={
                "endpoint_url": endpoint,
                "use_ssl": True,
            },
            config_kwargs={
                "s3": {"addressing_style": "path"}
            }
    )
    structure_folder = 'foldify_' + fasta_folder if structure_folder == '' else structure_folder
    remote_fastas = fs.ls(f"s3://{bucket}/{fasta_folder}", detail=True)
    remote_zips = fs.ls(f"s3://{bucket}/{structure_folder}", detail=True) # this is 10it/s; parallelize somehow?
    return remote_zips, remote_fastas


def ls(
    bucket = 'dmitrij',
    endpoint="https://8a6ab2cee54f34a71f5a8d99e92da2d2.r2.cloudflarestorage.com"
    ):
  fs = fsspec.filesystem(
            "s3",
            profile="r2",
            client_kwargs={
                "endpoint_url": endpoint,
                "use_ssl": True,
            },
            config_kwargs={
                "s3": {"addressing_style": "path"}
            }
  
    )
  remote = fs.ls(f"s3://{bucket}", detail=True)
  return ['/'.join(a['Key'].split('/')[1:]) for a in remote]


# make simple ls/rm command? 
# i.e., if logged in, we can just ls/rm as normal
# 

def ls_folder(  bucket = 'dmitrij',
    endpoint="https://8a6ab2cee54f34a71f5a8d99e92da2d2.r2.cloudflarestorage.com"):
    from natsort import natsorted
  

    folders = [folder for folder in ls(bucket=bucket, endpoint=endpoint) if '.' not in folder and len(folder) == 3]
    return natsorted(folders )


# merge two versions -- then rebuild foldifyai and redo plot in colab! 

def load(fasta_folder='000',
         structure_folder='',
         bucket='dmitrij',
         endpoint="https://8a6ab2cee54f34a71f5a8d99e92da2d2.r2.cloudflarestorage.com",
         use_local_cache=True,
         use_s3_cache=True,
         verbose=True,
         pbar=True,
         wipe_cache=False
        ):

    if structure_folder == '': structure_folder = 'foldify_' + fasta_folder
    if os.path.exists(f"{structure_folder}.csv") and use_local_cache:
        if verbose: print(f"[foldify] Found `{structure_folder}.csv` locally.")
        return pd.read_csv(f"{structure_folder}.csv")

    fs = fsspec.filesystem(
            "s3",
            profile="r2",
            client_kwargs={
                "endpoint_url": endpoint,
                "use_ssl": True,
            },
            config_kwargs={
                "s3": {"addressing_style": "path"}
            }

    )
    if fs.exists(f"s3://{bucket}/{structure_folder}.csv") and use_s3_cache:
        if verbose: print(f"[foldify] Found `{structure_folder}.csv` on s3.")
        fo = fsspec.open(
            f"{'' if wipe_cache else 'simplecache::'}s3://{bucket}/{structure_folder}.csv",
            mode="r",
            s3={"profile": "r2",
                "client_kwargs": {"endpoint_url": endpoint, "use_ssl": True},
                "config_kwargs": {"s3": {"addressing_style": "path"}},
                
                },
              skip_instance_cache=True
        )
        df = pd.read_csv(fo.open())
        df.to_csv(f"{structure_folder}.csv")
        return df

    # --- sync 000 (fasta files) ---
    remote = fs.ls(f"s3://{bucket}/{fasta_folder}", detail=True)
    os.makedirs(f"{fasta_folder}", exist_ok=True)
    for f in tqdm(remote, desc=f"s3 -> {fasta_folder}") if pbar else remote:
        local = f['Key'].replace(f'{bucket}/','')
        if not os.path.exists(local) or os.stat(local).st_size != f["Size"]:
            fs.get(f["Key"], str(local))

    # --- sync foldify_000 (zip files, then unzip) ---
    remote = fs.ls(f"s3://{bucket}/{structure_folder}", detail=True) # this is 10it/s; parallelize somehow?
    os.makedirs(f"{structure_folder}", exist_ok=True)
    for f in tqdm(remote, desc=f"s3 -> {structure_folder}") if pbar else remote:
        local = f['Key'].replace(f'{bucket}/', '')
        if not os.path.exists(local) or os.stat(local).st_size != f["Size"]:
            fs.get(f["Key"], str(local))

    # unzip local files
    files = [a for a in os.listdir(f"{structure_folder}") if a[-4:] == ".zip"]
    for local in tqdm(files, desc=f"{structure_folder} -> unzip") if pbar else files:
        if local.endswith(".zip") and not os.path.exists(f"{structure_folder}/{local[:-4]}"):
            shutil.unpack_archive(f"{structure_folder}/{local}", f"{structure_folder}/{local[:-4]}")

    fastas = os.listdir(f"{fasta_folder}")
    df = []
    for fasta in fastas:
        path1, path2 = f'{fasta_folder}/{fasta}', f'{structure_folder}/{fasta[:-6]}/boltz2_input.fasta'
        if not os.path.exists(path1): continue  # deal with this not existing when adding to dataframe
        if not os.path.exists(path2): continue
        lines1 = open(path1, 'r').read().split('\n')
        lines2 = open(path2, 'r').read().split('\n')
        seqs1, seqs2 = lines1[1::2], lines2[1::2]
        if not all([s1 == s2 for s1, s2 in zip(seqs1, seqs2)]):
            print( f"Sequences in {fasta} do not match boltz2_input.fasta\n" + str(seqs1) + "\n"+str(seqs2) + "\n" +\
                path1 + "\n" + path2)

        a3ms = [a for a in lines1 if ".a3m" in a]
        # todo check a3ms

        folder = fasta.replace('.fasta', '')
        path = f"{structure_folder}/{folder}/boltz2_confidence_0.json"
        if os.path.exists(path):
          dct = json.load(open(path, "r"))
          dct['filename']=fasta
          df.append(dct)
        else:
            pass # didn't finish this particular entry for some reason -- add to df somehow.


    df = pd.DataFrame(df)
    df.to_csv(f"{structure_folder}.csv")

    with fs.open(f"s3://{bucket}/{structure_folder}.csv", "w") as f:
        df.to_csv(f, index=False)

    return df

def ls(
  bucket = 'dmitrij',
  endpoint="https://8a6ab2cee54f34a71f5a8d99e92da2d2.r2.cloudflarestorage.com"
):
  fs = fsspec.filesystem(
            "s3",
            profile="r2",
            client_kwargs={
                "endpoint_url": endpoint,
                "use_ssl": True,
            },
            config_kwargs={
                "s3": {"addressing_style": "path"}
            }
  
    )
  remote = fs.ls(f"s3://{bucket}", detail=True)
  return ['/'.join(a['Key'].split('/')[1:]) for a in remote]

  
def upload_folder(
    folder,
    bucket="dmitrij",
    prefix="decile2",
    endpoint="https://8a6ab2cee54f34a71f5a8d99e92da2d2.r2.cloudflarestorage.com",
    verbose=True,
):
    fs = fsspec.filesystem(
        "s3",
        profile="r2",
        client_kwargs={"endpoint_url": endpoint, "use_ssl": True},
        config_kwargs={"s3": {"addressing_style": "path"}},
    )

    target_prefix = f"{bucket}/{prefix}/{os.path.basename(folder)}"

    for root, _, files in os.walk(folder):
        for file in tqdm(files):
            local_path = os.path.join(root, file)
            rel_path = os.path.relpath(local_path, folder)
            rel_path = '/'.join(folder.split('/')[-2:]) + local_path.split('/')[-1]
            remote_path = f"{target_prefix}{rel_path}"
            if verbose:
                print(f"Uploading {local_path} -> {remote_path}")
            
            fs.put_file(local_path, remote_path, asynchronous=True)




"""Command-line interface entrypoint for the `foldifyai` package."""
from __future__ import annotations
import time 
from colorama import Fore, Style
import fsspec

import base64
import zipfile
import io
import pathlib
import sys
from pathlib import Path
from tqdm import tqdm
import os 
import time 
import json 
from rdkit import Chem
from rdkit.Chem import AllChem
import urllib
from foldifyai.utils import get_type, file_exists
import requests
import fsspec

try: 
    from logmd import LogMD
except:
    pass 


def _usage() -> None:
    """Print a short help message using the actual executable name."""
    prog = pathlib.Path(sys.argv[0]).name or "foldify"
    print(f"Usage: {prog} <path_to_file.fasta>", file=sys.stderr)


def compute_3d_conformer(mol, version: str = "v3") -> bool:
    if version == "v3":
        options = AllChem.ETKDGv3()
    elif version == "v2":
        options = AllChem.ETKDGv2()
    else:
        options = AllChem.ETKDGv2()

    options.clearConfs = False
    conf_id = -1

    options.timeout = 3 # don't spend more than three seconds on AllChem.EmbedMolecule
    #options.maxIterations = 10 # don't spend more than 10 attempts (default is 100?)

    try:
        conf_id = AllChem.EmbedMolecule(mol, options)#, maxAttempts=0)

        if conf_id == -1:
            print(
                f"WARNING: RDKit ETKDGv3 failed to generate a conformer for molecule "
                f"{Chem.MolToSmiles(AllChem.RemoveHs(mol))}, so the program will start with random coordinates. "
                f"Note that the performance of the model under this behaviour was not tested."
            )
            options.useRandomCoords = True
            return False # conf_id = AllChem.EmbedMolecule(mol, options)

        #AllChem.UFFOptimizeMolecule(mol, confId=conf_id, maxIters=1000)
        # i set the maxIters=33 to skip more aggressively.
        AllChem.UFFOptimizeMolecule(mol, confId=conf_id, maxIters=33)

    except RuntimeError:
        return False 
        pass  # Force field issue here
    except ValueError:
        return False 
        pass  # sanitization issue here

    if conf_id != -1:
        conformer = mol.GetConformer(conf_id)
        conformer.SetProp("name", "Computed")
        conformer.SetProp("coord_generation", f"ETKDG{version}")
        return True

    return False

def test(seq, affinity=False):
    from pathlib import Path
    import hashlib

    cache_dir = Path.home() / ".cache"
    cache_dir.mkdir(parents=True, exist_ok=True)
    h = hashlib.sha256(seq.encode()).hexdigest()
    f = cache_dir / h
    if f.exists(): 
        print(f'hit ligand test cache {seq}')
        s = f.read_text()
        return "ok" in s, s  # cache mechanism 
    else: 
        print(f"didn't find ligand test cache {seq}, testing...")

    try:
        mol = AllChem.MolFromSmiles(seq)
        mol = AllChem.AddHs(mol)

        # Set atom names
        canonical_order = AllChem.CanonicalRankAtoms(mol)
        for atom, can_idx in zip(mol.GetAtoms(), canonical_order):
            atom_name = atom.GetSymbol().upper() + str(can_idx + 1)
            if len(atom_name) > 4:
                msg = (
                    f"{seq} has an atom with a name longer than "
                    f"4 characters: {atom_name}."
                )
                (cache_dir / h).write_text("fail\n" + msg)
                #raise ValueError(msg)
                return False, msg
            atom.SetProp("name", atom_name)

        success = compute_3d_conformer(mol)
        if not success:
            msg = f"Failed to compute 3D conformer for {seq}"
            (cache_dir / h).write_text("fail\n" + msg)
            return False, msg

        mol_no_h = AllChem.RemoveHs(mol, sanitize=False)
        affinity_mw = AllChem.Descriptors.MolWt(mol_no_h) if affinity else None
        (cache_dir / h).write_text("ok")
        return True, ""
    except Exception as e:
        print(e, seq)
        (cache_dir / h).write_text("fail\n" + str(e))
        return False, str(e)

def wait_on_server(host):
    d = {0: '-', 1: '/', 2: '\\', 3: '|', 4: '.'}
    i = 0 
    t0 = time.time()
    while True:
        url = f"{host}/ping"
        try:
            text = urllib.request.urlopen(url).read().decode()
            j = json.loads(text)
            if j['alive']: return
        except:
            t = time.time()-t0
            print(f'\rWaiting on server: {t:.2f}s {d[i%len(d)]}', end='')
            i+=1
            time.sleep(0.2)

LOCAL = 0
S3 = 1

def fold(args):
    folder = args.input
    log = args.logmd 

    if args.output == '': output_dir = f"foldify_{folder.replace('/','')}"
    else: output_dir = args.output

    s3_files = []
    mode = None 
    if args.s3 != '':
        # [ ] todo remove boto3
        '''import boto3
        s3 = boto3.session.Session(profile_name="r2").client('s3', endpoint_url=args.s3)
        resp = s3.list_objects_v2(Bucket='dmitrij', Prefix=output_dir)
        for obj in resp.get('Contents', []): 
            s3_files.append(obj['Key'].replace('.zip','.fasta').replace('foldify_', ''))

        if args.input[:5] != "s3://": 
            num_local = len(files)
            num_s3 = len(s3_files)

            _files = []
            num_overlap = 0
            for file in files: 
                if file not in s3_files: 
                    _files.append(file)
                else: 
                    num_overlap += 1
            files = _files

            print(f"Local finished: {num_local}")
            print(f"S3 finished: {num_s3} ")
            print(f"Overlap: {num_overlap}. ")
            print(f"Left: {len(files)}. ")'''

        fs = fsspec.filesystem(
            "s3",
            profile="r2",
            client_kwargs={
                "endpoint_url": args.s3,
                "use_ssl": True,
            },
            config_kwargs={
                "s3": {"addressing_style": "path"} 
            }
        )
        print("s3")
        mode = S3

        path = args.input[5:]
    else: 
        print("local")
        fs = fsspec.filesystem("file")
        path = args.input
        mode = LOCAL

    if args.input.endswith('.fasta'):
        files = [args.input]
        folder = folder.replace('.fasta', '')
    else:
        # [ ] port to s3 compatible. 
        #files = [a for a in Path(folder).rglob("*.fasta")]
        #files = sorted(files, key=lambda p: os.path.getsize(str(p)))
        #files = [str(a) for a in files]
        files = fs.glob(f"{path}*.fasta")
        files = sorted(files, key=lambda p: int(fs.size(p)))
        path = path.split('/')
        path = path[:1] + ['foldify_' + path[1]] + path[2:]
        path = "/".join(path)
        done = fs.glob(f"{path}*.zip")

        zip_ids = {x.split('/')[-1].rsplit('.', 1)[0] for x in done}

        if mode == S3: 
            left = [f"s3://{f}" for f in files if f.split('/')[-1].rsplit('.', 1)[0] not in zip_ids]
        elif mode == LOCAL:
            left = [f for f in files if f.split('/')[-1].rsplit('.', 1)[0] not in zip_ids]

        print("S3 jobs: ", len(files))
        print("S3 done: ", len(done))
        print("Left: ", len(left))

        files = left

        # if s3 don't look for local files.
        # but we want to look for stuff that's already finished. 

    if log: l = LogMD()


    wait_on_server(args.host)

    pbar = tqdm(files[::-1])
    for c,p in enumerate(pbar):

        # 
        if args.s3 != '': # handle single file vs folder. 
            pass 
            '''if "s3://" not in args.input: 
                s3_path = path[:-1] + '.zip'
                s3_path = '/'.join(s3_path.split('/')[0::2])
            else:
                s3_path = path[:-1] + '.zip'
                s3_path = '/'.join(s3_path.split('/')[0::2])'''
            s3_path = path + p.split('/')[-1].replace('.fasta', '.zip')
        else: 
            parts = p.split('/')
            middle = ["foldify_" + parts[-2]] if args.output == '' else [args.output]
            path = parts[:-2] + middle + [parts[-1].replace('.fasta', '')]
            path = "/".join(path) + '/'

            # add output dir to this. 

        #print(path)
        #exit()
        #print(path, s3_path)
        #exit()

        # [ ] s3 compatible. 
        if False: #os.path.exists(f"{path}boltz2_prediction_0.pdb") and not args.override:
            print(f"Local: Skipping {p} found {path}boltz2_prediction_0.pdb")

            # if args.s3 != '' it doesn't exist on s3 but does exist locally => upload it. 
            # this may have killed the data i already computed. 
            if args.s3 != '' and not file_exists(s3, 'dmitrij', s3_path):
                print(f"Uploading {p} to s3 {s3_path}")
                zip_bytes = io.BytesIO()
                with zipfile.ZipFile(zip_bytes, 'w') as zip_ref:
                    for root, dirs, files in os.walk(path):
                        for file in files:
                            zip_ref.write(os.path.join(root, file), os.path.relpath(os.path.join(root, file), path))
                zip_bytes.seek(0)
                with open('my_folder.zip', 'wb') as f: 
                   f.write(zip_bytes.getvalue())
                zip_bytes.seek(0)
                s3.upload_fileobj(zip_bytes, 'dmitrij', s3_path)
            continue 

        try: 
            skip = False 

            p = str(p)
            #content = open(p).read()
            content = fs.open(p, "rt").read()

            num_tokens = sum([len(line) for line in content.split('\n') if not line.startswith('>')])

            for line in content.split('\n'):
                if line.startswith('>'): continue 
                if line == '': continue 
                if get_type(line) == 'SMILES': 
                    if not test(line): 
                        print(f"Skipping {p}. RDKit didn't like {line}. ")
                        #open(new_path, 'w').write(f"Skipping {p}. RDKit didn't like {line}. ")
                        skip = True 
                    else: print('ok')
            if skip: continue 
            encoded = urllib.parse.quote(content, safe="")
            if len(content) == 0: 
                continue 

            url = f"{args.host}/fold?ui=False&only_return_zip=True&seq={encoded}&{args.args}&gpu={args.gpu}&get_msa_from_server={args.msa}"

            # Open connection with progress reporting
            response = urllib.request.urlopen(url)
            block_size = 1024
            
            result = ''
            while True:
                data = response.read(block_size)
                if not data:
                    break
                result += data.decode('utf-8')
                pbar.set_description(f"{Fore.BLUE}{Style.BRIGHT}[Foldify]{Style.RESET_ALL} {time.strftime('%d-%m %H:%M:%S')} {p} tokens={num_tokens} {len(result)/1000}KB")

            jsons = [json.loads(a) for a in result.split('\n@\n') if a != '']
            # last is a zip file, unzip. 
            b64_zip_data = jsons[-1]['data']

            zip_bytes = base64.b64decode(b64_zip_data)
            zip_in_memory = io.BytesIO(zip_bytes)
            with zipfile.ZipFile(zip_in_memory, 'r') as zip_ref:
                # [ ] remove s3 from this path 
                # this path didn't change in each iteration? 
                if mode == LOCAL: 
                    os.makedirs(path, exist_ok=True)
                    zip_ref.extractall(path) 

                #if args.s3 != '' and args.input[:5] != "s3://":
                elif mode == S3: 
                    with fs.open(f"{s3_path}", 'wb') as f:
                        f.write(zip_bytes)
                #    # integrate s3 with logmd -- parallel download s3 zips and store localcache? 
                '''elif args.s3 != '':

                    #import boto3
                    # https://8a6ab2cee54f34a71f5a8d99e92da2d2.r2.cloudflarestorage.com
                    #s3 = boto3.session.Session(profile_name="r2").client('s3', endpoint_url=args.s3)
                    #s3.upload_file('foldify_folders.zip', 'dmitrij', 'foldify_folders.zip')
                    zip_in_memory.seek(0)
                    s3.upload_fileobj(zip_in_memory, 'dmitrij', s3_path) # 'foldify_folders.zip')
                    print(s3_path)'''

            #time.sleep(1)

        except Exception as e: 
            print('something wrong', e)
            print(url)
            pass 
        #time.sleep(1)
        print('')
        wait_on_server(args.host)
        #exit()


def main() -> None:  # pragma: no cover

    import argparse

    parser = argparse.ArgumentParser(description='Foldify.ai CLI', add_help=False)
    parser.add_argument('-input', '-i', type=str, help='')
    parser.add_argument('-args', type=str, help='')
    parser.add_argument('-logmd', action='store_true', help='Log with LogMD')
    #parser.add_argument('-host', '-h', type=str, default='https://gpu1.foldify.org', help='Host URL for Foldify API')
    parser.add_argument('-host', '-h', type=str, default='http://0.0.0.0:8000', help='Host URL for Foldify API')
    parser.add_argument('-output', '-o', type=str, default='', help='Output directory for results')
    parser.add_argument('-gpu', '-g', type=int, default=0, help='GPU')
    parser.add_argument('-y', action='store_true', help='Pre-accept using remote host. ')
    parser.add_argument('-s3', type=str, default='', help='S3 endpoint (developed for cloudflared to lower cost).')
    parser.add_argument('-override', action='store_true', help='Override existing files')
    parser.add_argument('-msa', type=str, default='', help='Get msa from other ip. ')

    args = parser.parse_args()

    if args.host == 'https://gpu1.foldify.org' and not args.y:
        print("You didn't specify host. The default is a remote. ")
        print("Reply `REMOTE` if you want to send sequences. ")
        if input() != 'REMOTE': 
            print('Exiting.')
            exit()
        else: 
            print("Using remote host. ")
            print("You can skip this check with `foldify -y`")

    fold(args)


if __name__ == "__main__":  # pragma: no cover
    sys.argv = ['foldifyai','cofactors/']
    main() 



 
def get_type(seq: str) -> str:
    # normalise
    seq = seq.strip().upper().replace(" ", "").replace("\n", "")
    if not seq:
        raise ValueError("Empty sequence")

    # 1) quick SMILES test
    smiles_chars = set("[]=#@()+-\\/0123456789%")
    if any(c in smiles_chars for c in seq):
        return "SMILES"

    # 2) protein vs nucleic acid
    protein_letters = set("ACDEFGHIKLMNPQRSTVWYBXZJUO")   # AA + ambiguous codes
    dna_letters     = set("ACGTN")
    rna_letters     = set("ACGUN")

    letters_in_seq = set(seq)

    # anything outside nucleic-acid letters ⇒ protein
    if letters_in_seq - (dna_letters | rna_letters):
        return "protein"

    # only nucleic-acid letters left
    if 'U' in letters_in_seq and 'T' not in letters_in_seq:
        return "RNA"
    return "DNA"



def file_exists(s3_client, bucket, key):
    """ Checks if a file exists in an S3-compatible bucket.  """
    try:
        s3_client.head_object(Bucket=bucket, Key=key)
        return True
    except:
        return False